#include "opera1-fe.h"
#ifdef DVB_USB_LOG_PREFIX
#undef DVB_USB_LOG_PREFIX
#define DVB_USB_LOG_PREFIX "opera-s1-fe"
#endif
//#include <dvb/frontend.h>

#define LOF_HI			10600000
#define LOF_LO			9750000

struct opera1_fe_state {
	struct dvb_frontend fe;
	struct dvb_usb_device *d;

	fe_sec_voltage_t voltage;
	fe_sec_tone_mode_t tone_mode;

	u8 lnb_buf[8];

	u8 lock;
	u8 sig;
	u8 snr;

	//unsigned long next_status_check;
	//unsigned long status_check_interval;
};

struct opera1_request_type{
	int query;
	u8* config_buffer;
	struct dvb_usb_device *dev;
};

static struct dvb_frontend_ops opera1_fe_ops;

struct dvb_frontend * opera1_fe_attach(struct dvb_usb_device *d)
{
	struct opera1_fe_state *s = kmalloc(sizeof(struct opera1_fe_state), GFP_KERNEL);
	if (s == NULL)
		goto error;
	memset(s,0,sizeof(struct opera1_fe_state));

	s->d = d;
	s->fe.ops = &opera1_fe_ops;
	s->fe.demodulator_priv = s;

	//s->lnb_buf[1] = SET_LNB_POWER;
	s->lnb_buf[3] = 0xff; /* 0=tone burst, 2=data burst, ff=off */

	goto success;
error:
		return NULL;
success:
		return &s->fe;
}
static int opera_get_buffer(struct dvb_usb_device*dev, u8 type, u8 request, u8*data, u8 len){
	int value=0x0000;
	int ret;

	if ((ret=down_interruptible(&dev->usb_sem)))
		return ret;
	ret=usb_control_msg(dev->udev, usb_rcvctrlpipe(dev->udev,0),
		     type, USB_DIR_IN| USB_TYPE_VENDOR, request, value, data,len, 4000);
	
	up(&dev->usb_sem);
	return ret;
}


static void opera1_complete(struct urb* urb, struct pt_regs *regs){
	struct opera1_request_type *r=urb->context;
	u8 key[5];
	u8 rcbuffer[32];
	int retval;
	switch(urb->status) {
		case 0: /*success*/
			break;
		case -ECONNRESET:
		case -ENOENT:
		case -ESHUTDOWN:
		case -ETIMEDOUT:
			/* this urb is dead, cleanup */
			info("%s:urb shutting down with status: %d\n",
					__FUNCTION__, urb->status);
			
			usb_free_urb(urb);
			return;
		default:
			info("%s:nonzero status received: %d\n",
				__FUNCTION__,urb->status);
			goto exit;
	}
	
	
	if (r ){
		if (r->query==QUERY_RC && r->dev )
		{
			
			memset(rcbuffer,0,32);
			//info ("trying to get rc codes %08x",r->dev->udev);
			opera_get_buffer(r->dev, 0xb8, 0x00, rcbuffer, 32);
			
			/* TODO state, maybe it is VV ? */
			///if (key[1] != 0)
			///	key[0] = 0x01; /* if something is inside the buffer, simulate key press */
			///info("fb: %02x%02x%02x%02x%02x%02x%02x%02x",rcbuffer[0],rcbuffer[1],rcbuffer[2],rcbuffer[3],
			///     rcbuffer[4],rcbuffer[5],rcbuffer[6],rcbuffer[7]);
			/* call the universal NEC remote processor, to find out the key's state and event */
			///
			///dvb_usb_nec_rc_key_to_event(r->dev,key,/*event*/0,0/*state*/);
			///
			//if (key[0] != 0)
			//	deb_rc("key: %x %x %x %x %x\n",key[0],key[1],key[2],key[3],key[4]);
		}

		return ;
	}
	//usb_free_urb(u);
exit:	//retval = usb_submit_urb(urb, GFP_ATOMIC);
	usb_free_urb(urb);
	if (r ){
		if (r->config_buffer)
			kfree(r->config_buffer);
		kfree(r);
	}
	if(retval && 0)
		printk("%s - usb_free_urb failed with result: %d\n",
		       __FUNCTION__, retval);
}

int opera1_xilinx_send_async( struct dvb_usb_device*dev, u8 type, u8 thing, u8*data, u8 len,int req_type)
{
	
    	
    int ret;
	
	struct urb *u=usb_alloc_urb(0,GFP_KERNEL);
	struct opera1_request_type *r=kmalloc(sizeof(struct opera1_request_type),GFP_KERNEL);
	unsigned char* s=NULL;
	if ((ret=down_interruptible(&dev->usb_sem))){
		printk("%s: Failed to down usb semaphore.\n", __FUNCTION__);
		goto exit;
		//return ret;
	}
	if (u && r){
		
		memset(r,0,sizeof(struct opera1_request_type));
		s=kmalloc(9,GFP_KERNEL);
		s[0]=0x40;
		s[1]=type;
		s[2]=0;
		s[3]=thing;
		s[4]=0;
		s[5]=0;
		s[6]=0;
		s[7]=len;
		r->query=req_type;
		r->config_buffer=s;
		r->dev=dev;
		
		usb_fill_control_urb(u,dev->udev, usb_sndctrlpipe(dev->udev,0),s,
		data, len,opera1_complete,r);		
		usb_submit_urb(u,GFP_ATOMIC);
	}
	else
exit:
	{
		if (r)
			kfree(r);
		if (u)
			usb_free_urb(u);		
	}

	up(&dev->usb_sem);
	return 0;

	
}

int opera1_xilinx_send( struct dvb_usb_device*dev, u8 type, u8 thing, u8*data, u8 len)
{
	int value=0x0000;
	int ret;
	if ((ret=down_interruptible(&dev->usb_sem))){
	    info("sema was down");
	    
		return ret;
	}
	up(&dev->usb_sem);
	ret=usb_control_msg(dev->udev, usb_sndctrlpipe(dev->udev,0),
			       type, /*USB_DIR_OUT |*/ USB_TYPE_VENDOR, thing, value, data, len, 4000);
	//up(&dev->usb_sem);
	return ret;
	
}
static int opera1_fe_read_ber(struct dvb_frontend* fe, u32 *ber){
	
	int r=130;
	info("%s\n",__FUNCTION__);
	
	*ber = r << 8 | r;
	return 0;
}
static int opera1_fe_read_signal_strength(struct dvb_frontend* fe, u16 *strength){
	struct opera1_fe_state *st=fe->demodulator_priv;
	u8 r=0;
	int ber=155;
	if (st)
		ber=opera1_xilinx_send(st->d,0xb1,0xd1,&r,0);
	info("%s\n",__FUNCTION__);
	*strength=ber << 8 | ber;
	return 0;
}
static int opera1_fe_read_snr(struct dvb_frontend* fe, u16 *snr){
	int ber=155;
	info("%s\n",__FUNCTION__);
	*snr=ber << 8 | ber;
	return 0;
}

static int opera1_fe_set_frontend(struct dvb_frontend* fe,
																	struct dvb_frontend_parameters *fep)
{
	struct opera1_fe_state *st=fe->demodulator_priv;
	unsigned char freq_level[2],freq_set[4],symb_rate[4];
	int freq_add=0;
	u64 symb_rate_val=0;
	if (st){
		opera1_xilinx_send(st->d,0xb7,0x01,"\x01",1);
		opera1_xilinx_send(st->d,0xb1,0xd0,"\x05\x5b",2);
		opera1_xilinx_send(st->d,0xb1,0xc0,"\x07\xd0\xe1\xc2",4);
		opera1_xilinx_send(st->d,0xb1,0xd0,"\x05\x5b",2);
		opera1_xilinx_send(st->d,0xb1,0xc0,"\x07\xd0\xe5\xc2",4);
		opera1_xilinx_send(st->d,0xb1,0xd0,"\x05\x5b",2);
		opera1_xilinx_send(st->d,0xb1,0xc0,"\x07\xd0\xfd\xc6",4);
		opera1_xilinx_send(st->d,0xb1,0xd0,"\x05\x5b",2);
		opera1_xilinx_send(st->d,0xb1,0xd0,"\x08",1);
		opera1_xilinx_send(st->d,0xb1,0xd0,"\x00\xa1\x15\x00\x00\x7d\x05\x02\x00",9);
		opera1_xilinx_send(st->d,0xb1,0xd0,"\x0b\x00\x01\x81\x44\x19\x3f\x84\xda\x98\x95\xc9\xeb\x00\x19\x8b\x00\x82\x7f\x00\x00\x06\x50\x10\x00\x00\x37\xbc",28);
		switch (st->voltage){
			case SEC_VOLTAGE_13:
				opera1_xilinx_send(st->d,0xb6,0x01,"\x01",1);
				break;
			default:
				opera1_xilinx_send(st->d,0xb6,0x00,"\x00",1);
			
		}
		info("symbolrate set to %d",fep->u.qpsk.symbol_rate/1000);
		
		freq_level[0]=0x08;
		freq_level[1]=0x00;
		if (st->tone_mode==SEC_TONE_ON) //11700 Mhz ?
			freq_level[1]=0x03;
		symb_rate_val=fep->u.qpsk.symbol_rate;
		do_div(symb_rate_val,171875);
		symb_rate_val*=32768;
		//symb_rate_val=ep->u.qpsk.symbol_rate/1000/171875;//32768*fep->u.qpsk.symbol_rate/171875; //(125*1375);
		symb_rate[0]=0x1f;
		symb_rate[1]= (symb_rate_val>> 16) &0xff ;
		symb_rate[2]=(symb_rate_val >>8) &0xff ;
		symb_rate[3]=symb_rate_val &0xff ;
		info("symb set: %02X%02X%02X%02X",symb_rate[0],symb_rate[1],symb_rate[2],symb_rate[3]);
		opera1_xilinx_send(st->d,0xb1,0xd0,freq_level,2);
		opera1_xilinx_send(st->d,0xb1,0xd0,symb_rate,4);
		
		opera1_xilinx_send(st->d,0xb1,0xd0,"\x0f\x19",2);
		opera1_xilinx_send(st->d,0xb1,0xd0,"\x13\x98",2);
		opera1_xilinx_send(st->d,0xb1,0xd0,"\x14\x95",2);
		//set freq
		opera1_xilinx_send(st->d,0xb1,0xd0,"\x05\xb5",2);
		info("set freq to %d",(fep->frequency+((st->tone_mode==SEC_TONE_ON)?LOF_HI:LOF_LO))/1000);
		freq_add=fep->frequency/500;
		freq_set[0]=(freq_add>>8) & 0xff;
		freq_set[1]=freq_add & 0xff;
		freq_set[2]=0xe1;
		freq_set[3]=0xc0;
		info("freq set1: %02X%02X%02X%02X",freq_set[0],freq_set[1],freq_set[2],freq_set[3]);
		opera1_xilinx_send(st->d,0xb1,0xc0,freq_set,4);
		opera1_xilinx_send(st->d,0xb1,0xd0,"\x05\xb5",2);
		freq_set[2]=0xe5;
		
		info("freq set2: %02X%02X%02X%02X",freq_set[0],freq_set[1],freq_set[2],freq_set[3]);
		opera1_xilinx_send(st->d,0xb1,0xc0,freq_set,4);
		opera1_xilinx_send(st->d,0xb1,0xd0,"\x05\xb5",2);
		
		freq_set[2]=0xe5;
		freq_set[3]|=0x04;
		info("freq set3: %02X%02X%02X%02X",freq_set[0],freq_set[1],freq_set[2],freq_set[3]);
		opera1_xilinx_send(st->d,0xb1,0xc0,freq_set,4);
		opera1_xilinx_send(st->d,0xb1,0xd0,"\x05\xb5",2);

		opera1_xilinx_send(st->d,0xb1,0xd0,"\x05\xb5",2);
		opera1_xilinx_send(st->d,0xb1,0xd0,"\x0c\x00",2);
		opera1_xilinx_send(st->d,0xb1,0xd0,"\x13\x98",2);
		opera1_xilinx_send(st->d,0xb1,0xd0,"\x22\x00\x00",3);
		opera1_xilinx_send(st->d,0xb1,0xd0,"\x22",1);
		opera1_xilinx_send(st->d,0xb1,0xd0,"\x22",1);
		opera1_xilinx_send(st->d,0xb1,0xd0,"\x22",1);
		opera1_xilinx_send(st->d,0xb1,0xd0,"\x0c\x01",2);
		opera1_xilinx_send(st->d,0xb1,0xd0,"\x0c\x00",2);
		opera1_xilinx_send(st->d,0xb1,0xd0,"\x13",1);
		opera1_xilinx_send(st->d,0xb1,0xd0,"\x13\x98",2);
		
	}
	info("%s\n",__FUNCTION__);
	return 0;
}
static int opera1_fe_read_status(struct dvb_frontend* fe, fe_status_t *status)
{
	struct opera1_fe_state *st = fe->demodulator_priv;
	
	info("%s\n",__FUNCTION__);

	if (st->lock == 0)
		*status = FE_HAS_LOCK | FE_HAS_SYNC | FE_HAS_VITERBI | FE_HAS_SIGNAL | FE_HAS_CARRIER;
	else
		*status = 0;

	info("real state: %x\n",*status);
	*status = 0x1f;

	
	return 0;
}

static int opera1_fe_get_frontend(struct dvb_frontend* fe,
																	struct dvb_frontend_parameters *fep)
{
	info("%s\n",__FUNCTION__);
	return 0;
}
static void opera1_fe_release(struct dvb_frontend* fe)
{
	
	struct vp702x_fe_state *st = fe->demodulator_priv;
	info("%s\n",__FUNCTION__);
	if (st)
		kfree(st);
}
static int opera1_fe_set_voltage (struct dvb_frontend* fe, fe_sec_voltage_t
		voltage)
{
	struct opera1_fe_state *st = fe->demodulator_priv;
	
	info("%s\n",__FUNCTION__);

	st->voltage = voltage;
	return 0;
}
static int opera1_fe_set_tone(struct dvb_frontend* fe, fe_sec_tone_mode_t tone)
{
	struct opera1_fe_state *st = fe->demodulator_priv;
	
	info("%s tone=%d\n",__FUNCTION__,tone);
	if (st)
		st->tone_mode=tone;
	return 0;
}


static struct dvb_frontend_ops opera1_fe_ops = {
	.info = {
		.name           = "DVB-Opera S1 frontend DVB-S",
		.type           = FE_QPSK,
		.frequency_min       = 950000,
		.frequency_max       = 2150000,
		.frequency_stepsize  = 1000,   /* kHz for QPSK frontends */
		.frequency_tolerance = 0,
		.symbol_rate_min     = 1000000,
		.symbol_rate_max     = 45000000,
		.symbol_rate_tolerance = 500,  /* ppm */
		.caps = FE_CAN_FEC_1_2 | FE_CAN_FEC_2_3 | FE_CAN_FEC_3_4 |
				FE_CAN_FEC_5_6 | FE_CAN_FEC_7_8 |
				FE_CAN_QPSK |
				FE_CAN_FEC_AUTO
	},
	.release = opera1_fe_release,

	.init = NULL,
	.sleep = NULL,

	.set_frontend = opera1_fe_set_frontend,
	.get_frontend = opera1_fe_get_frontend,
	/*.get_tune_settings = vp702x_fe_get_tune_settings,*/

	.read_status = opera1_fe_read_status,
	.read_ber = opera1_fe_read_ber,
	.read_signal_strength = opera1_fe_read_signal_strength,
	.read_snr = opera1_fe_read_snr,
	/*.read_ucblocks = vp702x_fe_read_unc_blocks,

	.diseqc_send_master_cmd = vp702x_fe_send_diseqc_msg,
	.diseqc_send_burst = vp702x_fe_send_diseqc_burst,*/
	.set_tone = opera1_fe_set_tone,
	.set_voltage = opera1_fe_set_voltage,
};

